import os
from pathlib import Path
from sys import _getframe
from typing import Iterator
import json
from logging import getLogger

logger = getLogger(__name__)


class Config:
    """A class for handling configuration settings for the aiomql package.

    Keyword Args:
        **kwargs: Configuration settings as keyword arguments.
        Variables set this way supersede those set in the config file.

    Attributes:
        record_trades (bool): Whether to keep record of trades or not.
        filename (str): Name of the config file
        records_dir (str): Path to the directory where trade records are saved
        win_percentage (float): Percentage of achieved target profit in a trade to be considered a win
        login (int): Trading account number
        password (str): Trading account password
        server (str): Broker server
        path (str): Path to terminal file
        timeout (int): Timeout for terminal connection

    Notes:
        By default, the config class looks for a file named aiomql.json.
        You can change this by passing the filename keyword argument to the constructor.
        By passing reload=True to the load_config method, you can reload and search again for the config file.
    """
    login: int = 0
    password: str = ''
    server: str = ''
    path: str = ''
    timeout: int = 60000
    record_trades: bool = True
    filename: str = 'aiomql.json'
    win_percentage: float = 0.85
    records_dir = Path.home() / 'Documents' / 'Aiomql' / 'Trade Records' if record_trades else None
    _load = 1

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, '_instance'):
            cls._instance = super().__new__(cls)
        return cls._instance
    
    def __init__(self, **kwargs):
        self.load_config(reload=False)
        [setattr(self, key, value) for key, value in kwargs]

        
    @staticmethod
    def walk_to_root(path: str) -> Iterator[str]:
        
        if not os.path.exists(path):
            raise IOError('Starting path not found')
        
        if os.path.isfile(path):
            path = os.path.dirname(path)
        
        last_dir = None
        current_dir = os.path.abspath(path)
        while last_dir != current_dir:
            yield current_dir
            parent_dir = os.path.abspath(os.path.join(current_dir, os.path.pardir))
            last_dir, current_dir = current_dir, parent_dir
    
    def find_config(self):
        current_file = __file__
        frame = _getframe()
        while frame.f_code.co_filename == current_file:
            if frame.f_back is None:
                return None
            frame = frame.f_back
        frame_filename = frame.f_code.co_filename
        path = os.path.dirname(os.path.abspath(frame_filename))
        
        for dirname in self.walk_to_root(path):
            check_path = os.path.join(dirname, self.filename)
            if os.path.isfile(check_path):
                return check_path
        return None
    
    def load_config(self, file: str = None, reload: bool = True):
        if reload:
            self._load = 1
        if self._load != 1:
            return

        self._load = 0
        data = {}
        if (file := (file or self.find_config())) is None:
            logger.warning('No Config File Found')
        else:
            fh = open(file, mode='r')
            data = json.load(fh)
            fh.close()
        [setattr(self, key, value) for key, value in data.items()]
        self.records_dir.mkdir(parents=True, exist_ok=True) if self.records_dir else ...

    def account_info(self) -> dict['login', 'password', 'server']:
        """Returns Account login details as found in the config object if available

           Returns:
               dict: A dictionary of login details
        """
        return {'login': self.login, 'password': self.password, 'server': self.server}



from aiomql import config

from . import get_config, get_default_config


def test_default_config_file(get_default_config):
    conf = config.Config()
    assert conf.win_percentage == 0.90


def test_config_file_name(get_config):
    conf = config.Config(filename='config.json')
    assert conf.win_percentage == 0.8



from aiomql.symbol import Symbol

from . import *


class TestSymbol(BaseTest):
    sym = Symbol(name="EURJPY")

    async def test_init(self):
        await self.sym.init()
        assert self.sym.select is True



from ...symbol import Symbol
from ...core.exceptions import VolumeError


class ForexSymbol(Symbol):
    """Subclass of Symbol for Forex Symbols. Handles the conversion of currency and the computation of stop loss,
    take profit and volume.
    """
    @property
    def pip(self):
        """Returns the pip value of the symbol. This is ten times the point value for forex symbols.

        Returns:
            float: The pip value of the symbol.
        """
        return self.point * 10

    async def compute_volume(self, *, amount: float, pips: float, use_limits: bool = False) -> float:
        """Compute volume given an amount to risk and target pips. Round the computed volume to the nearest step.

        Args:
            amount (float): Amount to risk. Given in terms of the account currency.
            pips (float): Target pips.
            use_limits (bool): If True, the computed volume checked against the maximum and minimum volume.

        Returns:
            float: volume

        Raises:
            VolumeError: If the computed volume is less than the minimum volume or greater than the maximum volume.
        """
        volume = amount / (self.pip * pips)
        volume = self.round_off_volume(volume)
        if self.check_volume(volume)[0]:
            return volume
        if use_limits:
            return self.check_volume(volume)[1]
        raise VolumeError(f'Incorrect Volume. Computed Volume outside the range of permitted volumes')



from . import *
from aiomql import Terminal


class TestTerminal(BaseTest):
    terminal = Terminal()
    async def test_version(self):
        res = await self.terminal.version
        assert len(res) == 3

    async def test_info(self):
        res = await self.terminal.info()
        assert res.connected is True

    async def test_error(self):
        res = await self.terminal.last_error()
        assert res.code == 1

    async def test_symbols_get(self):
        res = await self.terminal.symbols_get()
        sym = next(res)
        assert isinstance(sym.name, str)



"""Utility functions for aiomql."""
def dict_to_string(data: dict, multi=False) -> str:
    """Convert a dict to a string. Use for logging.

    Args:
        data (dict): The dict to convert.
        multi (bool, optional): If True, each key-value pair will be on a new line. Defaults to False.

    Returns:
        str: The string representation of the dict.
    """
    sep = '\n' if multi else ', '
    return f"{sep}".join(f"{key}: {value}\n" for key, value in data.items())


import asyncio
import logging
from typing import Literal
from dataclasses import dataclass

from ...symbol import Symbol
from ...trader import Trader
from ...candle import Candles
from ...strategy import Strategy
from ...core import TimeFrame, OrderType
from ...sessions import Sessions

logger = logging.getLogger(__name__)


@dataclass
class Entry:
    """
    Entry class for FingerTrap strategy. Will be used to store entry conditions and other entry related data.

    Attributes:
        bearish (bool): True if the market is bearish
        bullish (bool): True if the market is bullish
        ranging (bool): True if the market is ranging
        snooze (float): Time to wait before checking for entry conditions
        trend (str): The current trend of the market
        new (bool): True if the last candle is new
        order_type (OrderType): The type of order to place
        pips (int): The number of pips to place the order from the current price
    """
    bearish: bool = False
    bullish: bool = False
    ranging: bool = True
    trending: bool = False
    trend: Literal['ranging', 'bullish', 'bearish'] = 'ranging'
    snooze: float = 0
    last_trend_time: float = 0
    last_entry_time: float = 0
    new: bool = True
    order_type: OrderType | None = None
    pips: float = 0

    def update(self, **kwargs):
        fields = self.__dict__
        for key in kwargs:
            if key in fields:
                setattr(self, key, kwargs[key])
        match self.trend:
            case 'ranging':
                self.ranging = True
                self.trending = self.bullish = self.bearish = False
            case 'bullish':
                self.ranging = self.bearish = False
                self.bullish = self.trending = True
            case 'bearish':
                self.ranging = self.bullish = False
                self.bearish = self.trending = True


class FingerTrap(Strategy):
    trend_time_frame: TimeFrame
    entry_time_frame: TimeFrame
    trend: int
    fast_period: int
    slow_period: int
    entry_period: int
    parameters: dict
    prices: Candles
    name = "FingerTrap"
    pips: float
    interval: TimeFrame
    entry_candles_count: int
    trend_candles_count: int

    def __init__(self, *, symbol: Symbol, params: dict | None = None, trader: Trader = None, sessions: Sessions = None):
        super().__init__(symbol=symbol, params=params, sessions=sessions)
        self.trend = self.parameters.get('trend', 3)
        self.fast_period = self.parameters.setdefault('fast_period', 8)
        self.slow_period = self.parameters.setdefault('slow_period', 34)
        self.entry_time_frame = self.parameters.setdefault('entry_time_frame', TimeFrame.M5)
        self.trend_time_frame = self.parameters.setdefault('trend_time_frame', TimeFrame.H1)
        self.trader = trader or Trader(symbol=self.symbol)
        self.pips = self.parameters.setdefault('pips', 10)
        self.entry: Entry = Entry(snooze=self.trend_time_frame.time, pips=self.pips)
        self.entry_period = self.parameters.setdefault('entry_period', 8)

        self.trend_candles_count = self.parameters.setdefault('trend_candles_count',
                                                              86400 // self.trend_time_frame.time)
        self.trend_candles_count = max(self.trend_candles_count, self.slow_period)
        self.entry_candles_count = self.trend_candles_count * (self.trend_time_frame.time // self.entry_time_frame.time)
        self.entry_candles_count = max(self.entry_candles_count, self.entry_period)

    async def check_trend(self):
        try:
            candles = await self.symbol.copy_rates_from_pos(timeframe=self.trend_time_frame,
                                                            count=self.trend_candles_count)
            current = candles[-1]
            if current.time > self.entry.last_trend_time:
                self.entry.update(new=True, last_trend_time=current.time)
            else:
                self.entry.update(new=False)
                return

            candles.ta.ema(length=self.slow_period, append=True, fillna=0)
            candles.ta.ema(length=self.fast_period, append=True, fillna=0)
            candles.rename(inplace=True, **{f'EMA_{self.fast_period}': 'fast', f'EMA_{self.slow_period}': 'slow'})

            # Compute
            candles['fast_A_slow'] = candles.ta_lib.above(candles.fast, candles.slow)
            candles['fast_B_slow'] = candles.ta_lib.below(candles.fast, candles.slow)
            candles['close_A_fast'] = candles.ta_lib.above(candles.close, candles.fast)
            candles['close_B_fast'] = candles.ta_lib.below(candles.close, candles.fast)

            trend = candles[-self.trend: -1]
            if all((c.is_bullish() and c.fast_A_slow and c.close_A_fast) for c in trend):
                self.entry.update(trend='bullish')

            elif all(c.is_bearish() and c.fast_B_slow and c.close_B_fast for c in trend):
                self.entry.update(trend='bearish')

            else:
                self.entry.update(trend='ranging', snooze=self.trend_time_frame.time)
        except Exception as exe:
            logger.error(f'{exe}. Error in {self.__class__.__name__}.check_trend')

    async def confirm_trend(self):
        try:
            candles = await self.symbol.copy_rates_from_pos(timeframe=self.entry_time_frame,
                                                            count=self.entry_candles_count)
            current = candles[-1]
            if current.time > self.entry.last_entry_time:
                self.entry.update(new=True, last_entry_time=current.time)
            else:
                self.entry.update(new=False)
                return

            candles.ta.ema(length=self.entry_period, append=True, fillna=0)
            candles.rename(**{f'EMA_{self.entry_period}': 'ema'})
            candles['close_XA_ema'] = candles.ta_lib.cross(candles.close, candles.ema)
            candles['close_XB_ema'] = candles.ta_lib.cross(candles.close, candles.ema, above=False)
            current = candles[-2]

            if self.entry.bullish and current.close_XA_ema:
                self.entry.update(snooze=self.entry_time_frame.time, order_type=OrderType.BUY)
            elif self.entry.bearish and current.close_XB_ema:
                self.entry.update(snooze=self.entry_time_frame.time, order_type=OrderType.SELL)
            else:
                self.entry.update(snooze=self.entry_time_frame.time, order_type=None)
        except Exception as exe:
            logger.error(f'{exe} Error in {self.__class__.__name__}.confirm_trend')

    async def watch_market(self):
        await self.check_trend()
        if not self.entry.ranging:
            await self.confirm_trend()

    async def trade(self):
        logger.info(f'Trading {self.symbol}')
        async with self.sessions as sess:
            while True:
                await sess.check()
                try:
                    await self.watch_market()
                    if not self.entry.new:
                        await asyncio.sleep(2)
                        continue

                    if self.entry.order_type is None:
                        await self.sleep(self.entry.snooze)
                        continue

                    await self.trader.place_trade(order_type=self.entry.order_type, pips=self.entry.pips,
                                                  params=self.parameters)
                    await self.sleep(self.entry.snooze)
                except Exception as err:
                    logger.error(f"Error: {err}\t Symbol: {self.symbol} in {self.__class__.__name__}.trade")
                    await self.sleep(self.trend_time_frame.time)
                    continue



"""Risk Assessment and Management"""
from .account import Account
from .symbol import Symbol


class RAM:
    account: Account = Account()
    risk_to_reward: float
    risk: float
    amount: float
    pips: float
    volume: float

    def __init__(self, *, risk_to_reward: float = 1, risk: float = 0.01, amount: float = 0, pips: float = 0, volume=0):
        """Initialize Risk Assessment and Management with the provided keyword arguments.

        Keyword Args:
            risk_to_reward (float): Risk to reward ratio. Defaults to 1
            risk (float): Percentage of account balance to risk per trade 0.01 # 1%
            amount (float): Amount to risk per trade in terms of account currency 0
            pips (float): Target pips to risk
            volume (float): Volume to trade 0
        """
        self.risk_to_reward = risk_to_reward
        self.risk = risk
        self.amount = amount
        self.pips = pips
        self.volume = volume

    async def get_amount(self, risk: float = 0) -> float:
        """Calculate the amount to risk per trade as a percentage of equity.

        Keyword Args:
            risk (float): Percentage of account balance to risk per trade. Defaults to zero.

        Returns:
            float: Amount to risk per trade
        """
        await self.account.refresh()
        risk = risk or self.risk
        return self.account.equity * risk

    async def get_volume(self, *, symbol: Symbol, pips: float = 0, amount: float = 0) -> float:
        """Calculate the volume to trade. if pips is not provided, the pips attribute is used.
        If the amount attribute or amount argument is zero, the amount is calculated using the get_amount method based
        on the risk.

        Keyword Args:
            symbol (Symbol): Financial instrument
            pips (float): Target pips. Defaults to zero.
            amount (float): Amount to risk per trade. Defaults to zero.

        Returns:
            float: Volume to trade
        """
        pips = pips or self.pips
        amount = amount or self.amount or await self.get_amount()
        return await symbol.compute_volume(amount=amount, pips=pips)



"""Trader class module. Handles the creation of an order and the placing of trades"""

from datetime import datetime
from typing import TypeVar
from logging import getLogger
from zoneinfo import ZoneInfo

from .order import Order
from .symbol import Symbol as _Symbol
from .ram import RAM
from .core.models import OrderType
from .core.config import Config
from .utils import dict_to_string
from .result import Result

logger = getLogger(__name__)
Symbol = TypeVar('Symbol', bound=_Symbol)


class Trader:
    """Base class for creating a Trader object. Handles the creation of an order and the placing of trades

    Attributes:
        symbol (Symbol): Financial instrument class Symbol class or any subclass of it.
        ram (RAM): RAM instance
        order (Order): Trade order

    Class Attributes:
        name (str): A name for the strategy.
        account (Account): Account instance.
        mt5 (MetaTrader): MetaTrader instance.
        config (Config): Config instance.
    """
    config = Config()

    def __init__(self, *, symbol: Symbol, ram: RAM = None):
        """Initializes the order object and RAM instance

        Args:
            symbol (Symbol): Financial instrument
            ram (RAM): Risk Assessment and Management instance
        """
        self.symbol = symbol
        self.order = Order(symbol=symbol.name)
        self.ram = ram or RAM()

    async def create_order(self, *, order_type: OrderType, **kwargs):
        """Complete the order object with the required values. Creates a simple order.
        Uses the ram instance to set the volume.

        Args:
            order_type (OrderType): Type of order
            kwargs: keyword arguments as required for the specific trader
        """
        # check if pips is passed in as a keyword argument, if not use the pips attribute of the ram instance
        pips = kwargs.get('pips', 0) or self.ram.pips
        self.order.volume = kwargs.get('volume', self.ram.volume) or await self.ram.get_volume(symbol=self.symbol,
                                                                                               pips=pips)
        self.order.type = order_type
        await self.set_order_limits(pips=pips)

    async def set_order_limits(self, pips: float):
        """Sets the stop loss and take profit for the order.
        This method uses pips as defined for forex instruments.

        Args:
            pips: Target pips
        """
        # use passed in pips and the pip value of the symbol to calculate the stop loss and take profit.
        # this is sure to work for forex instruments.
        pips = pips * self.symbol.pip
        sl, tp = pips, pips * self.ram.risk_to_reward
        tick = await self.symbol.info_tick()
        if self.order.type == OrderType.BUY:
            self.order.sl, self.order.tp = tick.ask - sl, tick.ask + tp
            self.order.price = tick.ask
        else:
            self.order.sl, self.order.tp = tick.bid + sl, tick.bid - tp
            self.order.price = tick.bid

    async def place_trade(self, order_type: OrderType, params: dict = None, **kwargs):
        """Places a trade based on the order_type.

        Args:
            order_type (OrderType): Type of order
            params: parameters of the trading strategy used to place the trade
            kwargs: keyword arguments as required for the specific trader
        """
        try:
            await self.create_order(order_type=order_type, **kwargs)

            # Check the order before placing it
            check = await self.order.check()
            if check.retcode != 0:
                logger.warning(
                    f"Symbol: {self.order.symbol}\nResult:\n{dict_to_string(check.get_dict(include={'comment', 'retcode'}), multi=True)}")
                return

            # check expected profit
            profit = await self.order.calc_profit()

            # Send the order.
            result = await self.order.send()
            if result.retcode != 10009:
                logger.warning(
                    f"Symbol: {self.order.symbol}\nResult:\n{dict_to_string(result.get_dict(include={'comment', 'retcode'}), multi=True)}")
                return

            logger.info(f"Symbol: {self.order.symbol}\nOrder: {dict_to_string(result.dict, multi=True)}\n")

            # save trade result and passed in parameters
            if result.retcode == 10009 and self.config.record_trades:
                params = params or {}
                params['expected_profit'] = profit
                date = datetime.utcnow()
                date = date.replace(tzinfo=ZoneInfo('UTC'))
                params['date'] = date
                params['time'] = date.timestamp()
                res = Result(result=result, parameters=params)
                await res.save_csv()
        except Exception as err:
            logger.error(f"{err}. Symbol: {self.order.symbol}\n {self.__class__.__name__}.place_trade")
